; ---------------------------------------------------------------------------
; PT3 Turbo Sound player for Amstrad CPC
; (c) 2013 SyX
; Based in the Universal PT2'n'PT3 Turbo Sound player for ZX Spectrum
; (c)2004-2007 S.V.Bulba <vorobey@mail.khstu.ru>
; http//bulba.untergrund.net/ (http//bulba.at.kz/)
; and the Amstrad CPC/Plus Port by Grim/Arkos^Semilanceata 
; 2014 modified by Prodatron for Maxam assembler (WinApe) and SymAmp/SymbOS
; 2014 modified by Shining for SDCC
; ---------------------------------------------------------------------------


; Defines for the conditional assembly

PSG_TEST        	.equ 0  ; psg testmode
CPC_FREQ         	.equ 0	; Use CPC frequencies or ZX
DISABLE_INTERRUPT 	.equ 1	; 0: Player will be called within irq ; 1: Player handles ei/di itself
							; Also the Play-Routine only saves all Registers (except AF') when player handles ei/di


; Global vars for use in C
.globl _PC_Stop
.globl _PC_Init
.globl _PC_Play
.globl _PC_Get_Volume


_PC_Init:
	push af
	push bc
	push DE
	push HL	
	push IX
	push IY
	
	; Set ZX PSG frequency in the YMZs
	LD   BC, #0xF880
	LD   HL, #0x7704
	OUT  (C),H                      ; Set Timer 0 as clock for the YMZs
	OUT  (C),L                      ; Set the clock to 1,75 MHz (ZX frequency)
	
	call _TS_Init ;Init the song
	
	pop IY
	pop IX
	pop HL
	pop DE
	pop BC
	pop af
	ret 
	
_PC_Play:
.if   DISABLE_INTERRUPT
	push af
	push bc
	push DE
	push HL	
	push IX
	push IY
.endif   DISABLE_INTERRUPT
	call _TS_Play
.if   DISABLE_INTERRUPT	
	pop IY
	pop IX
	pop HL
	pop DE
	pop BC
	pop af
.endif   DISABLE_INTERRUPT	
	ret 	




; Features
; --------
; - Can be compiled at any address (i.e. no need rounding ORG
;    address).
; - Variables (VARS) can be located at any address (not only after
;    code block).
; - TS_Init subprogram checks PT3-module version and rightly
;    generates both note and volume tables outside of code block
;    (in VARS).
; - Two portamento (spc. command 3xxx) algorithms (depending of
;    PT3 module version).
; - New 1.XX and 2.XX special command behaviour (only for PT v3.7
;    and higher).
; - Any Tempo value are accepted (including Tempo=1 and Tempo=2).
; - TS modes 2xPT3, 2xPT2 and PT v3.7 TS standard.
; - Fully compatible with Ay_Emul PT3 and PT2 players codes.

; Limitations
; -----------
; - Can run in RAM only (self-modified code is used).
; - PT2 position list must be end by #FF marker only.

; Warning!!! _TS_Play subprogram can crash if no module are loaded
; into RAM or TS_Init subprogram was not called before.

; Call TS_Stop or TS_Init one more time to mute sound after stopping
; playing 

; Constants with the AY/YM registers 
TonA        .EQU 0            ; Channel A Frequency
TonB        .EQU 2            ; Channel B Frequency
TonC        .EQU 4            ; Channel C Frequency
Noise       .EQU 6            ; Noise Frequency 
Mixer       .EQU 7            ; PSG/YM Control 
AmplA       .EQU 8            ; Channel A Amplitude
AmplB       .EQU 9            ; Channel B Amplitude
AmplC       .EQU 10           ; Channel C Amplitude
Env         .EQU 11           ; Envelope Frequency
EnvTp       .EQU 13           ; Envelope Shape/Cycle

TS_SETUP:    .DB 0  ; set bit0, if you want to play without looping (optional)
                    ; set bit1 for PT2 and reset for PT3 before calling TS_Init
                    ; bits2-3 %00-ABC, %01-ACB, %10-BAC (optional)
                    ; bits4-5 %00-no TS, %01-2 modules TS, %10-
                    ;           autodetect PT3 TS-format by AlCo (PT 3.7+);
                    ;           Remark old PT3 TS-format by AlCo (PT 3.6) is not
                    ;           documented and must be converted to new standard.
                    ; bit6 is set each time, when loop point of 2nd TS
                    ;          module is passed (optional).
                    ; bit7 is set each time, when loop point of 1st TS
                    ;          or of single module is passed (optional).

CHECKLP:
    LD   HL,#TS_SETUP
    BIT  0,VRS_ModNum-100(IY)
    JR   Z,CHL1
    SET  6,(HL)
    JR   CHL2
CHL1:
    SET  7,(HL)
CHL2:
    BIT  0,(HL)
    RET  Z
    POP  HL
    INC  VRS_DelyCnt-100(IY)
    INC  VRS_ChanA+CHP_NtSkCn-100(IY)
    XOR  A
    LD   VRS_AYREGS+AmplA-100(IY),A
    LD   VRS_AYREGS+AmplB-100(IY),A
    LD   VRS_AYREGS+AmplC-100(IY),A
    RET

; ---------------------------------------------------------------------------
; Stop the music and silent the channels.
; ---------------------------------------------------------------------------
_PC_Stop:
    XOR  A
    LD   H,A
    LD   L,A
    LD   (#VARS1+VRS_AYREGS+AmplA),A
    LD   (#VARS1+VRS_AYREGS+AmplB),HL
    LD   (#VARS2+VRS_AYREGS+AmplA),A
    LD   (#VARS2+VRS_AYREGS+AmplB),HL
    JP   TS_Play_ROUT
	
	
_PC_Get_Volume:

	LD IX,#2
	ADD IX,SP	
	LD A, 0( IX ) ; CHIP
	CP #1
	ld hl, #VARS2+VRS_AYREGS+AmplA
	jr z,PC_Get_Volume_Chip_2
	
	ld hl, #VARS1+VRS_AYREGS+AmplA
	
PC_Get_Volume_Chip_2:	
	LD A, 1( IX )
	
	ld b,#0
	ld c,a
	add hl,bc
	ld A,(hl)
	and #0xF
	ld L,A
	ret



; ---------------------------------------------------------------------------
; Initialize the song
; ENTRIES
;     HL = Address of module or first module in case of split songs
;     DE = Address of second module in case of split songs (optional)
; ---------------------------------------------------------------------------
_TS_Init:
	LD IX,#16
	ADD IX,SP
	LD L, 0( IX )
    LD H, 1( IX )
	LD E, 2( IX )
    LD D, 3( IX )
    PUSH DE
    PUSH HL
	
	; Initialize Song
    LD   A,#0b00010000    ; 2 x PT3 TS | ABC | Loop Song
    LD   (TS_SETUP),A   ; Set the TS_SETUP variable
    
    ; Clean VARS1 + VARS2
    LD   HL,#VARS
    LD   (HL),#0
    LD   DE,#VARS+1
    LD   BC,#VAR0END-VARS-1
    LDIR
 
    INC  HL
    LD   (#VARS1+VRS_AdInPtA),HL ; ptr to zero
    LD   (#VARS2+VRS_AdInPtA),HL

    POP  HL
    LD   IY,#VARS1+100
    CALL INITPT3

    POP  HL
    ; Use version and tone table of 1st module *** REVIEW for reduce size in files ***
    LD   A,13-100(IX)  ; Extract version number
    SUB  #0x30            ; A - '0'
    JR   C,TS_Init_L20
    CP   #10             ; < 10?
    JR   C,TS_Init_L21
TS_Init_L20:
    LD   A,#6            ; Version 6 for unknown type
TS_Init_L21:
    LD   (Version),A    ; Set version for the song

    ; Test for Turbo-Sound
    LD   IY,#VARS2+100
    LD   A,(TS_SETUP)
    AND  #0b00110000      ; Mask bits 5-4
    RET  Z              ; (JR Z,NOTS) %00 = No Turbo Sound
    CP   #16
    JR   Z,#TwoPT3s      ; %01 = 2 x PT3 songs
    LD   A,(#Version)
    CP   #7
    RET  C              ; (JR C,NOTS) Song version <> 7, then no Turbo Sound
    LD   A,98-100(IX)  ; Get Alco Turbo Sound marker
    CP   #32             ; is space?
    RET  Z              ; (JR Z,NOTS) Yes, then NO Turbo Sound
    ; It's a Turbo Sound song, copy VARS1 to VARS2
    LD   HL,#VARS1
    LD   DE,#VARS2
    LD   BC,#VRS_datlen
    LDIR

    SET  1,VRS_ModNum-100(IY)
    LD   C,A
    ADD  A,A
    ADD  A,C
    SUB  #2
    LD   (TSSub),A
    JR   AlCoTS_
TwoPT3s:
    CALL INITPT3
AlCoTS_:
    LD   A,#1
    LD   (is_ts),A
    SET  0,VRS_ModNum-100(IY)

NOTS:
    LD   A,#1
    LD   (#VARS1+VRS_DelyCnt),A
    LD   (#VARS2+VRS_DelyCnt),A

    LD   HL,#0xF001       ; H - CHP_Volume, L - CHP_NtSkCn
    LD   (#VARS1+VRS_ChanA+CHP_NtSkCn),HL
    LD   (#VARS1+VRS_ChanB+CHP_NtSkCn),HL
    LD   (#VARS1+VRS_ChanC+CHP_NtSkCn),HL
    LD   (#VARS2+VRS_ChanA+CHP_NtSkCn),HL
    LD   (#VARS2+VRS_ChanB+CHP_NtSkCn),HL
    LD   (#VARS2+VRS_ChanC+CHP_NtSkCn),HL

    LD   HL,#PT3EMPTYORN
    LD   (#VARS1+VRS_ChanA+CHP_OrnPtr),HL
    LD   (#VARS1+VRS_ChanB+CHP_OrnPtr),HL
    LD   (#VARS1+VRS_ChanC+CHP_OrnPtr),HL
    LD   (#VARS2+VRS_ChanA+CHP_OrnPtr),HL
    LD   (#VARS2+VRS_ChanB+CHP_OrnPtr),HL
    LD   (#VARS2+VRS_ChanC+CHP_OrnPtr),HL

    JP   TS_Play_ROUT

    ; Initialize variables for PT3 song
INITPT3:
    ; Set song pointer
    LD   VRS_MODADDR-100(IY),L
    LD   VRS_MODADDR+1-100(IY),H

    PUSH HL
    LD   DE,#100
    ADD  HL,DE
    LD   A,(HL)             ; Get song Tempo
    LD   VRS_Delay-100(IY),A
    PUSH HL
    POP  IX
    ADD  HL,DE

    ; SETCPPT
    LD   VRS_CrPsPtr-100(IY),L
    LD   VRS_CrPsPtr+1-100(IY),H

    LD   E,102-100(IX)
    INC  HL

    LD   VRS_PosSub-100(IY),L

    ADD  HL,DE
    ; SETLPPT
    LD   VRS_LPosPtr-100(IY),L
    LD   VRS_LPosPtr+1-100(IY),H
    POP  DE
    LD   L,103-100(IX)
    LD   H,104-100(IX)
    ADD  HL,DE
    ; SETPTPT
    LD   VRS_PatsPtr-100(IY),L
    LD   VRS_PatsPtr+1-100(IY),H
    LD   HL,#169
    ADD  HL,DE
    ; SETORPT
    LD   VRS_OrnPtrs-100(IY),L
    LD   VRS_OrnPtrs+1-100(IY),H
    LD   HL,#105
    ADD  HL,DE
    ; SETSMPT
    LD   VRS_SamPtrs-100(IY),L
    LD   VRS_SamPtrs+1-100(IY),H
    RET

PTDECOD:
    PUSH IY
    POP  IX
    ADD  IX,DE
    PUSH IY
    POP  IX
    ADD  IX,DE
    JP   PT3PD

; PT3 pattern decoder
PD_OrSm:
    LD   CHP_Env_En-12(IX),#0
    CALL SETORN
PD_SAM_:
    LD   A,(BC)
    INC  BC
    RRCA

PD_SAM:
    CALL SETSAM
    JR   PD_LOOP

PD_VOL:
    RRCA
    RRCA
    RRCA
    RRCA
    LD   CHP_Volume-12(IX),A
    JR   PD_LP2
    
PD_EOff:
    LD   CHP_Env_En-12(IX),A
    LD   CHP_PsInOr-12(IX),A
    JR   PD_LP2

PD_SorE:
    DEC  A
    JR   NZ,PD_ENV
    LD   A,(BC)
    INC  BC
    LD   CHP_NNtSkp-12(IX),A
    JR   PD_LP2

PD_ENV:
    CALL SETENV
    JR   PD_LP2

PD_ORN:
    CALL SETORN
    JR   PD_LOOP

PD_ESAM:
    LD   CHP_Env_En-12(IX),A
    LD   CHP_PsInOr-12(IX),A
    CALL NZ,SETENV
    JR   PD_SAM_

PT3PD:
    LD   A,CHP_Note-12(IX)
    LD   (PrNote+1),A
    LD   L,CHP_CrTnSl-12(IX)
    LD   H,CHP_CrTnSl+1-12(IX)
    LD   (PrSlide+1),HL

PD_LOOP:
    LD   DE,#0x2010
PD_LP2:
    LD   A,(BC)
    INC  BC
    ADD  A,E
    JR   C,PD_OrSm
    ADD  A,D
    JR   Z,PD_FIN
    JR   C,PD_SAM
    ADD  A,E
    JR   Z,PD_REL
    JR   C,PD_VOL
    ADD  A,E
    JR   Z,PD_EOff
    JR   C,PD_SorE
    ADD  A,#96
    JR   C,PD_NOTE
    ADD  A,E
    JR   C,PD_ORN
    ADD  A,D
    JR   C,PD_NOIS
    ADD  A,E
    JR   C,PD_ESAM
    ADD  A,A
    LD   E,A
    LD   HL,#SPCCOMS+0xFF20-0x2000
    ADD  HL,DE
    LD   E,(HL)
    INC  HL
    LD   D,(HL)
    PUSH DE
    JR   PD_LOOP

PD_NOIS:
    LD   VRS_Ns_Base-100(IY),A
    JR   PD_LP2

PD_REL:
    RES  0,CHP_Flags-12(IX)
    JR   PD_RES

PD_NOTE:
    LD   CHP_Note-12(IX),A
    SET  0,CHP_Flags-12(IX)
    XOR  A

PD_RES:
.if   DISABLE_INTERRUPT
    DI                  ; Disable Ints for SP
.endif ;DISABLE_INTERRUPT
    LD   (PDSP_ + 1),SP
    LD   SP,IX
    LD   H,A
    LD   L,A
    PUSH HL
    PUSH HL
    PUSH HL
    PUSH HL
    PUSH HL
    PUSH HL
PDSP_:
    LD   SP,#0x0000
    .if   DISABLE_INTERRUPT
    EI                  ; Reenable Ints 
	.endif ;DISABLE_INTERRUPT

PD_FIN:
    LD   A,CHP_NNtSkp-12(IX)
    LD   CHP_NtSkCn-12(IX),A
    RET

C_PORTM:
    LD   A,(BC)
    INC  BC
    ; Skip precalculated tone delta (because cannot be right after pt3 compilation)
    INC  BC
    INC  BC
    EX   AF,AF'
    LD   A,(BC)         ; Signed tone step
    INC  BC
    LD   (LoStep),A
    LD   A,(BC)
    INC  BC
    AND  A
    EX   AF,AF'
    LD   L,CHP_CrTnSl-12(IX)
    LD   H,CHP_CrTnSl+1-12(IX)

; Set portamento variables
; A - Delay; A' - Hi(Step); ZF' - (A'=0); HL - CrTnSl
SETPORT:
    RES  2,CHP_Flags-12(IX)
    LD   CHP_TnSlDl-12(IX),A
    LD   CHP_TSlCnt-12(IX),A
    PUSH HL
    LD   DE,#NT_
    LD   A,CHP_Note-12(IX)
    LD   CHP_SlToNt-12(IX),A
    ADD  A,A
    LD   L,A
    LD   H,#0
    ADD  HL,DE
    LD   A,(HL)
    INC  HL
    LD   H,(HL)
    LD   L,A
    PUSH HL
PrNote:
    LD   A,#0x3E
    LD   CHP_Note-12(IX),A
    ADD  A,A
    LD   L,A
    LD   H,#0
    ADD  HL,DE
    LD   E,(HL)
    INC  HL
    LD   D,(HL)
    POP  HL
    SBC  HL,DE
    LD   CHP_TnDelt-12(IX),L
    LD   CHP_TnDelt+1-12(IX),H
    POP  DE
VersLabel:	
Version  .EQU VersLabel+1
    LD   A,#0x3E
    CP   #6
    JR   C,OLDPRTM  ; Old 3xxx for PT v3.5- *** Review ***
PrSlide:
    LD   DE,#0x1111
    LD   CHP_CrTnSl-12(IX),E
    LD   CHP_CrTnSl+1-12(IX),D
LoStepLabel:
LoStep   .EQU LoStepLabel+1
OLDPRTM:
    LD   A,#0x3E
    EX   AF,AF'
    JR   Z,NOSIG
    EX   DE,HL
NOSIG:
    SBC  HL,DE
    JP   P,SET_STP
    CPL
    EX   AF,AF'
    NEG
    EX   AF,AF'
SET_STP:
    LD   CHP_TSlStp+1-12(IX),A
    EX   AF,AF'
    LD   CHP_TSlStp-12(IX),A
    LD   CHP_COnOff-12(IX),#0
    RET

C_GLISS:
    SET  2,CHP_Flags-12(IX)
    LD   A,(BC)
    INC  BC
    LD   CHP_TnSlDl-12(IX),A
    AND  A
    JR   NZ,GL36
    LD   A,(Version)    ; AlCo PT3.7+
    CP   #7
    SBC  A,A
    INC  A
GL36:
    LD   CHP_TSlCnt-12(IX),A
    LD   A,(BC)
    INC  BC
    EX   AF,AF'
    LD   A,(BC)
    INC  BC
    JR   SET_STP

C_SMPOS:
    LD   A,(BC)
    INC  BC
    LD   CHP_PsInSm-12(IX),A
    RET

C_ORPOS:
    LD   A,(BC)
    INC  BC
    LD   CHP_PsInOr-12(IX),A
    RET

C_VIBRT:
    LD   A,(BC)
    INC  BC
    LD   CHP_OnOffD-12(IX),A
    LD   CHP_COnOff-12(IX),A
    LD   A,(BC)
    INC  BC
    LD   CHP_OffOnD-12(IX),A
    XOR  A
    LD   CHP_TSlCnt-12(IX),A
    LD   CHP_CrTnSl-12(IX),A
    LD   CHP_CrTnSl+1-12(IX),A
    RET

C_ENGLS:
    LD   A,(BC)
    INC  BC
    LD   VRS_Env_Del-100(IY),A
    LD   VRS_CurEDel-100(IY),A
    LD   A,(BC)
    INC  BC
    LD   L,A
    LD   A,(BC)
    INC  BC
    LD   H,A
    LD   VRS_ESldAdd-100(IY),L
    LD   VRS_ESldAdd+1-100(IY),H
    RET

C_DELAY:
    LD   A,(BC)
    INC  BC
    LD   VRS_Delay-100(IY),A
    LD   HL,#VARS2+VRS_ModNum    ; if AlCo_TS
    BIT  1,(HL)
    RET  Z
    LD   (VARS1+VRS_Delay),A
    LD   (VARS1+VRS_DelyCnt),A
    LD   (VARS2+VRS_Delay),A
    RET
    
SETENV:
    LD   CHP_Env_En-12(IX),E
    LD   VRS_AYREGS+EnvTp-100(IY),A
    LD   A,(BC)
    INC  BC
    LD   H,A
    LD   A,(BC)
    INC  BC
    LD   L,A
    ; SETENBS
    LD   VRS_EnvBase-100(IY),L
    LD   VRS_EnvBase+1-100(IY),H
    XOR  A
    LD   CHP_PsInOr-12(IX),A
    LD   VRS_CurEDel-100(IY),A
    LD   H,A
    LD   L,A
    ; SETESLD
    LD   VRS_CurESld-100(IY),L
    LD   VRS_CurESld+1-100(IY),H
    RET

SETORN:
    ADD  A,A
    LD   E,A
    LD   D,#0
    LD   CHP_PsInOr-12(IX),D
    LD   L,VRS_OrnPtrs-100(IY)
    LD   H,VRS_OrnPtrs+1-100(IY)
    ADD  HL,DE
    LD   E,(HL)
    INC  HL
    LD   D,(HL)
    LD   L,VRS_MODADDR-100(IY)
    LD   H,VRS_MODADDR+1-100(IY)
    ADD  HL,DE
    LD   CHP_OrnPtr-12(IX),L
    LD   CHP_OrnPtr+1-12(IX),H
C_NOP:
    RET

SETSAM:
    ADD  A,A
    LD   E,A
    LD   D,#0
    LD   L,VRS_SamPtrs-100(IY)
    LD   H,VRS_SamPtrs+1-100(IY)
    ADD  HL,DE
    LD   E,(HL)
    INC  HL
    LD   D,(HL)
    LD   L,VRS_MODADDR-100(IY)
    LD   H,VRS_MODADDR+1-100(IY)
    ADD  HL,DE
    LD   CHP_SamPtr-12(IX),L
    LD   CHP_SamPtr+1-12(IX),H
    RET

; All 16 addresses to protect from broken pt3 modules
SPCCOMS:
    .DW C_NOP
    .DW C_GLISS
    .DW C_PORTM
    .DW C_SMPOS
    .DW C_ORPOS
    .DW C_VIBRT
    .DW C_NOP
    .DW C_NOP
    .DW C_ENGLS
    .DW C_DELAY
    .DW C_NOP
    .DW C_NOP
    .DW C_NOP
    .DW C_NOP
    .DW C_NOP
    .DW C_NOP

CHREGS:
    PUSH IY
    POP  IX
    ADD  IX,DE
    XOR  A
    LD   (Ampl),A
    BIT  0,CHP_Flags(IX)
    PUSH HL
    JP   Z,CH_EXIT
    .if   DISABLE_INTERRUPT
    DI                  ; Disable Ints for SP
	.endif ;DISABLE_INTERRUPT
    LD   (CSP_+1),SP
    LD   L,CHP_OrnPtr(IX)
    LD   H,CHP_OrnPtr+1(IX)
    LD   SP,HL
    POP  DE
    LD   H,A
    LD   A,CHP_PsInOr(IX)
    LD   L,A
    ADD  HL,SP
    INC  A
OrnCP:
    CP   D
    JR   C,CH_ORPS
OrnLD:
    LD   A,E
CH_ORPS:
    LD   CHP_PsInOr(IX),A
    LD   A,CHP_Note(IX)
    ADD  A,(HL)
    JP   P,CH_NTP
    XOR  A
CH_NTP:
    CP   #96
    JR   C,CH_NOK
    LD   A,#95
CH_NOK:
    ADD  A,A
    EX   AF,AF'
    LD   L,CHP_SamPtr(IX)
    LD   H,CHP_SamPtr+1(IX)
    LD   SP,HL
    POP  DE
    LD   H,#0
    LD   A,CHP_PsInSm(IX)
    LD   B,A
    ADD  A,A
SamClc2:
    ADD  A,A
    LD   L,A
    ADD  HL,SP
    LD   SP,HL
    LD   A,B
    INC  A
SamCP:
    CP   D
    JR   C,CH_SMPS
SamLD:
    LD   A,E
CH_SMPS:
    LD   CHP_PsInSm(IX),A
    POP  BC
    POP  HL
    LD   E,CHP_TnAcc(IX)
    LD   D,CHP_TnAcc+1(IX)
    ADD  HL,DE
    BIT  6,B
    JR   Z,CH_NOAC
    LD   CHP_TnAcc(IX),L
    LD   CHP_TnAcc+1(IX),H
CH_NOAC:
    EX   DE,HL
    EX   AF,AF'

    ld hl,#NT_
    add l
    ld l,a
    adc h
    sub l
    ld h,a

    LD   SP,HL
    POP  HL
    ADD  HL,DE
    LD   E,CHP_CrTnSl(IX)
    LD   D,CHP_CrTnSl+1(IX)
    ADD  HL,DE
CSP_:
    LD   SP,#0x0000
    .if   DISABLE_INTERRUPT
    EI                  ; Reenable Ints 
	.endif ;DISABLE_INTERRUPT
    EX   (SP),HL
    XOR  A
    OR   CHP_TSlCnt(IX)
    JR   Z,CH_AMP
    DEC  CHP_TSlCnt(IX)
    JR   NZ,CH_AMP
    LD   A,CHP_TnSlDl(IX)
    LD   CHP_TSlCnt(IX),A
    LD   L,CHP_TSlStp(IX)
    LD   H,CHP_TSlStp+1(IX)
    LD   A,H
    ADD  HL,DE
    LD   CHP_CrTnSl(IX),L
    LD   CHP_CrTnSl+1(IX),H
    BIT  2,CHP_Flags(IX)
    JR   NZ,CH_AMP
    LD   E,CHP_TnDelt(IX)
    LD   D,CHP_TnDelt+1(IX)
    AND  A
    JR   Z,CH_STPP
    EX   DE,HL
CH_STPP:
    SBC  HL,DE
    JP   M,CH_AMP
    LD   A,CHP_SlToNt(IX)
    LD   CHP_Note(IX),A
    XOR  A
    LD   CHP_TSlCnt(IX),A
    LD   CHP_CrTnSl(IX),A
    LD   CHP_CrTnSl+1(IX),A
CH_AMP:
    LD   A,CHP_CrAmSl(IX)
    BIT  7,C
    JR   Z,CH_NOAM
    BIT  6,C
    JR   Z,CH_AMIN
    CP   #15
    JR   Z,CH_NOAM
    INC  A
    JR   CH_SVAM
CH_AMIN:
    CP   #-15
    JR   Z,CH_NOAM
    DEC  A
CH_SVAM:
    LD   CHP_CrAmSl(IX),A
CH_NOAM:
    LD   L,A
    LD   A,B
    AND  #15
    ADD  A,L
    JP   P,CH_APOS
    XOR  A
CH_APOS:
    CP   #16
    JR   C,CH_VOL
    LD   A,#15
CH_VOL:
    OR   CHP_Volume(IX)

    push de
    ld de,#VT_
    ADD  A,e
    LD   L,A
    ADC  A,d
    SUB  L
    LD   H,A
    pop de

    LD   A,(HL)
CH_ENV:
    BIT  0,C
    JR   NZ,CH_NOEN
    OR   CHP_Env_En(IX)
CH_NOEN:
    LD   (Ampl),A
    BIT  7,B
    LD   A,C
    JR   Z,NO_ENSL
    RLA
    RLA
    SRA  A
    SRA  A
    SRA  A
    ADD  A,CHP_CrEnSl(IX)      ; See comment below
    BIT  5,B
    JR   Z,NO_ENAC
    LD   CHP_CrEnSl(IX),A
NO_ENAC:
    ADD  A,VRS_AddToEn-100(IY) ; Bug in PT3 - Need word here
    LD   VRS_AddToEn-100(IY),A
    JR   CH_MIX
NO_ENSL:
    RRA
    ADD  A,CHP_CrNsSl(IX)
    LD   VRS_AddToNs-100(IY),A
    BIT  5,B
    JR   Z,CH_MIX
    LD   CHP_CrNsSl(IX),A
CH_MIX:
    LD   A,B
    RRA
    AND  #0x48
CH_EXIT:
    OR   VRS_AYREGS+Mixer-100(IY)
    RRCA
    LD   VRS_AYREGS+Mixer-100(IY),A
    POP  HL
    XOR  A
    OR   CHP_COnOff(IX)
    RET  Z
    DEC  CHP_COnOff(IX)
    RET  NZ
    XOR  CHP_Flags(IX)
    LD   CHP_Flags(IX),A
    RRA
    LD   A,CHP_OnOffD(IX)
    JR   C,CH_ONDL
    LD   A,CHP_OffOnD(IX)
CH_ONDL:
    LD   CHP_COnOff(IX),A
    RET

PLAY_:
    XOR  A
    LD   VRS_AddToEn-100(IY),A
    LD   VRS_AYREGS+Mixer-100(IY),A
    DEC  A
    LD   VRS_AYREGS+EnvTp-100(IY),A
    DEC  VRS_DelyCnt-100(IY)
    JP   NZ,PL2
    DEC  VRS_ChanA+CHP_NtSkCn-100(IY)
    JR   NZ,PL1B
    LD   C,VRS_AdInPtA-100(IY)
    LD   B,VRS_AdInPtA+1-100(IY)
    LD   A,(BC)
    AND  A
    JR   NZ,PL1A
    LD   D,A
    LD   VRS_Ns_Base-100(IY),A
    LD   L,VRS_CrPsPtr-100(IY)
    LD   H,VRS_CrPsPtr+1-100(IY)
    INC  HL
    LD   A,(HL)
    INC  A
    JR   NZ,PLNLP

    CALL CHECKLP

    LD   L,VRS_LPosPtr-100(IY)
    LD   H,VRS_LPosPtr+1-100(IY)
    LD   A,(HL)
    INC  A
PLNLP:
    ; SETCPPT
    LD   VRS_CrPsPtr-100(IY),L
    LD   VRS_CrPsPtr+1-100(IY),H
    DEC  A
    BIT  1,VRS_ModNum-100(IY)
    JR   Z,NoAlCo
TSSubLabel:	
TSSub    .EQU TSSubLabel+1
    SUB  #0xD6
    CPL
NoAlCo:
    ADD  A,A
    LD   E,A
    RL   D

    LD   A,L
    SUB  VRS_PosSub-100(IY)
    LD   VRS_CurPos-100(IY),A

    LD   L,VRS_PatsPtr-100(IY)
    LD   H,VRS_PatsPtr+1-100(IY)
    ADD  HL,DE
    LD   E,VRS_MODADDR-100(IY)
    LD   D,VRS_MODADDR+1-100(IY)
    .if   DISABLE_INTERRUPT
    DI                  ; Disable Ints for SP
	.endif ;DISABLE_INTERRUPT
    LD   (PSP_+1),SP
    LD   SP,HL
    POP  HL
    ADD  HL,DE
    LD   B,H
    LD   C,L
    POP  HL
    ADD  HL,DE
    LD   VRS_AdInPtB-100(IY),L
    LD   VRS_AdInPtB+1-100(IY),H
    POP  HL
    ADD  HL,DE
    LD   VRS_AdInPtC-100(IY),L
    LD   VRS_AdInPtC+1-100(IY),H
PSP_:
    LD   SP,#0000
    .if   DISABLE_INTERRUPT
    EI                  ; Reenable Ints 
	.endif ;DISABLE_INTERRUPT
PL1A:
    LD   DE,#VRS_ChanA+12-100
    CALL PTDECOD
    LD   VRS_AdInPtA-100(IY),C
    LD   VRS_AdInPtA+1-100(IY),B

PL1B:
    DEC  VRS_ChanB+CHP_NtSkCn-100(IY)
    JR   NZ,PL1C
    LD   DE,#VRS_ChanB+12-100
    LD   C,VRS_AdInPtB-100(IY)
    LD   B,VRS_AdInPtB+1-100(IY)
    CALL PTDECOD
    LD   VRS_AdInPtB-100(IY),C
    LD   VRS_AdInPtB+1-100(IY),B

PL1C:
    DEC  VRS_ChanC+CHP_NtSkCn-100(IY)
    JR   NZ,PL1D
    LD   DE,#VRS_ChanC+12-100
    LD   C,VRS_AdInPtC-100(IY)
    LD   B,VRS_AdInPtC+1-100(IY)
    CALL PTDECOD
    LD   VRS_AdInPtC-100(IY),C
    LD   VRS_AdInPtC+1-100(IY),B

PL1D:
    LD   A,VRS_Delay-100(IY)
    LD   VRS_DelyCnt-100(IY),A

PL2:
    LD   DE,#VRS_ChanA-100
    LD   L,VRS_AYREGS+TonA-100(IY)
    LD   H,VRS_AYREGS+TonA+1-100(IY)
    CALL CHREGS
    LD   VRS_AYREGS+TonA-100(IY),L
    LD   VRS_AYREGS+TonA+1-100(IY),H
AmplLabel:
Ampl     .EQU AmplLabel+1
    LD   A,#0x3E
    LD   VRS_AYREGS+AmplA-100(IY),A
    LD   DE,#VRS_ChanB-100
    LD   L,VRS_AYREGS+TonB-100(IY)
    LD   H,VRS_AYREGS+TonB+1-100(IY)
    CALL CHREGS
    LD   VRS_AYREGS+TonB-100(IY),L
    LD   VRS_AYREGS+TonB+1-100(IY),H
    LD   A,(Ampl)
    LD   VRS_AYREGS+AmplB-100(IY),A
    LD   DE,#VRS_ChanC-100
    LD   L,VRS_AYREGS+TonC-100(IY)
    LD   H,VRS_AYREGS+TonC+1-100(IY)
    CALL CHREGS
    LD   VRS_AYREGS+TonC-100(IY),L
    LD   VRS_AYREGS+TonC+1-100(IY),H
    LD   A,(Ampl)
    LD   VRS_AYREGS+AmplC-100(IY),A

    LD   A,VRS_Ns_Base-100(IY)
    ADD  VRS_AddToNs-100(IY)
    LD   VRS_AYREGS+Noise-100(IY),A

    LD   A, VRS_AddToEn-100(IY)
    LD   E,A
    ADD  A,A
    SBC  A,A
    LD   D,A
    LD   L, VRS_EnvBase-100(IY)
    LD   H, VRS_EnvBase+1-100(IY)
    ADD  HL,DE
    LD   E, VRS_CurESld-100(IY)
    LD   D, VRS_CurESld+1-100(IY)
    ADD  HL,DE
    LD   VRS_AYREGS+Env-100(IY),L
    LD   VRS_AYREGS+Env+1-100(IY),H

    XOR  A
    OR   VRS_CurEDel-100(IY)
    RET  Z
    DEC  VRS_CurEDel-100(IY)
    RET  NZ
    LD   A, VRS_Env_Del-100(IY)
    LD   VRS_CurEDel-100(IY),A
    LD   L, VRS_ESldAdd-100(IY)
    LD   H, VRS_ESldAdd-100+1(IY)
    ADD  HL,DE
    ; SETESLD
    ;LD   (IY-#100+#VRS_CurESld),L
	LD   VRS_CurESld-100(IY),L
    ;LD   (IY-#100+#VRS_CurESld+#1),H
	LD VRS_CurESld-100+1(iy),h
    RET

; ---------------------------------------------------------------------------
; Play sound
; ---------------------------------------------------------------------------
_TS_Play:
    ; Proccess Music
    ; 1.- YM LEFT
    LD   IY,#VARS1+100
    CALL PLAY_

    ; 2.- PSG
    LD   IY,#VARS2+100
    CALL PLAY_

TS_Play_ROUT:
.IF PSG_TEST
        LD HL,VARS1+VRS_AYREGS        ;*** CPC SOUNDOUTPUT ***
        xor a

LOUT1:  call writepsghl
        inc a
        cp 13
        jr nz,LOUT1
        bit 7,(hl)
        ret nz

;; A = register
;; (HL) = value
writepsghl:
        ld b,#f4                    ; [2]
        out (c),A                   ; [4]
        ld bc,#f6c0                 ; [3]
        out (c),c                   ; [4]
        defb #ed                    ; [4] 
        defb #71                    ; out (c),0
        ld b,#f5                    ; [2]
        outi                        ; [5]
        ;;out (c),e                   ; [4]
        ld bc,#f680                 ; [3]
        out (c),c                   ; [4]
        defb #ed                    ; [4] 
        defb #71                    ; out (c),0
                                    ; = [34]
        ret

.ELSE
    ; Send datas to the audio chips
TS_Play_ROUT_YM_LEFT:
    ; 1.- YMZ Left
    LD   HL,#VARS1+VRS_AYREGS
    LD   BC,#0xF988

    XOR  A                      ; YM Register 0
    ; Send Registers 0 - 12

    OUT (C),A
	OUTI
	INC B
	INC A
    OUT (C),A
	OUTI
	INC B
	INC A
    OUT (C),A
	OUTI
	INC B
	INC A
    OUT (C),A
	OUTI
	INC B
	INC A
    OUT (C),A
	OUTI
	INC B
	INC A
    OUT (C),A
	OUTI
	INC B
	INC A
    OUT (C),A
	OUTI
	INC B
	INC A
    OUT (C),A
	OUTI
	INC B
	INC A
    OUT (C),A
	OUTI
	INC B
	INC A
    OUT (C),A
	OUTI
	INC B
	INC A
    OUT (C),A
	OUTI
	INC B
	INC A
    OUT (C),A
	OUTI
	INC B
	INC A
    OUT (C),A
	OUTI
	INC B
	INC A

    ; Send Register 13 (Envelope Shape/Cycle)
    OUT  (C),A
    LD   A,(HL)
    AND  A
    JP   M,TS_Play_ROUT_YM_RIGHT       ; RET     M
    OUTI                        ; Value

TS_Play_ROUT_YM_RIGHT:
    ; 2.- YM Right
    LD   HL,#VARS2+VRS_AYREGS
    LD   BC,#0xF984

    XOR  A                      ; YM Register 0
    ; Send Registers 0 - 12

    OUT (C),A
	OUTI
	INC B
	INC A
    OUT (C),A
	OUTI
	INC B
	INC A
    OUT (C),A
	OUTI
	INC B
	INC A
    OUT (C),A
	OUTI
	INC B
	INC A
    OUT (C),A
	OUTI
	INC B
	INC A
    OUT (C),A
	OUTI
	INC B
	INC A
    OUT (C),A
	OUTI
	INC B
	INC A
    OUT (C),A
	OUTI
	INC B
	INC A
    OUT (C),A
	OUTI
	INC B
	INC A
    OUT (C),A
	OUTI
	INC B
	INC A
    OUT (C),A
	OUTI
	INC B
	INC A
    OUT (C),A
	OUTI
	INC B
	INC A
    OUT (C),A
	OUTI
	INC B
	INC A

    ; Send Register 13 (Envelope Shape/Cycle)
    OUT  (C),A
    LD   A,(HL)
    AND  A
    RET  M
    OUTI                        ; Value

    RET
.ENDIF
; ---------------------------------------------------------------------------
PT3EMPTLABEL:
PT3EMPTYORN .EQU PT3EMPTLABEL-1
    .DB 1,0

; vars from here can be stripped (you can move VARS to any other address)
VARS:
is_ts:   .DB 0          ; 0 = No Turbo Sound | 1 = Turbo Sound

; ChannelsVars

; reset group
CHP_PsInOr  .equ  0       ;DEFB 0
CHP_PsInSm  .equ  1       ;DEFB 0
CHP_CrAmSl  .equ  2       ;DEFB 0
CHP_CrNsSl  .equ  3       ;DEFB 0
CHP_CrEnSl  .equ  4       ;DEFB 0
CHP_TSlCnt  .equ  5       ;DEFB 0
CHP_CrTnSl  .equ  6       ;DEFW 0
CHP_TnAcc   .equ  8       ;DEFW 0
CHP_COnOff  .equ 10       ;DEFB 0
; reset group

CHP_OnOffD  .equ 11       ;DEFB 0

; IX for PTDECOD here (+12)
CHP_OffOnD  .equ 12       ;DEFB 0
CHP_OrnPtr  .equ 13       ;DEFW 0
CHP_SamPtr  .equ 15       ;DEFW 0
CHP_NNtSkp  .equ 17       ;DEFB 0
CHP_Note    .equ 18       ;DEFB 0
CHP_SlToNt  .equ 19       ;DEFB 0
CHP_Env_En  .equ 20       ;DEFB 0
CHP_Flags   .equ 21       ;DEFB 0
; Enabled - 0, SimpleGliss - 2
CHP_TnSlDl  .equ 22       ;DEFB 0
CHP_TSlStp  .equ 23       ;DEFW 0
CHP_TnDelt  .equ 25       ;DEFW 0
CHP_NtSkCn  .equ 27       ;DEFB 0
CHP_Volume  .equ 28       ;DEFB 0

CHP_datlen  .equ 29


VRS_ModNum  .equ  0       ;DEFB 0          ; bit0 ChipNum
                         ;                  bit1 1-reversed patterns order (AlCo TS)

VRS_ChanA   .equ 0*29+1   ;DEFS CHP_datlen
VRS_ChanB   .equ 1*29+1   ;DEFS CHP_datlen
VRS_ChanC   .equ 2*29+1   ;DEFS CHP_datlen

; GlobalVars
VRS_MODADDR .equ  88 ; 88      ;DEFW 0          ; Pointer to the song
VRS_OrnPtrs .equ  90 ; 90      ;DEFW 0
VRS_SamPtrs .equ  92 ; 92      ;DEFW 0
VRS_PatsPtr .equ  94 ; 94      ;DEFW 0
VRS_AdInPtA .equ  96 ; 96      ;DEFW 0
VRS_AdInPtB .equ  98 ; 98      ;DEFW 0
VRS_AdInPtC .equ 100 ;100      ;DEFW 0
VRS_CrPsPtr .equ 102 ;102      ;DEFW 0
VRS_LPosPtr .equ 104 ;104      ;DEFW 0
                     ;
VRS_Delay   .equ 106 ;106      ;DEFB 0          ; Song Tempo
VRS_DelyCnt .equ 107 ;107      ;DEFB 0
VRS_ESldAdd .equ 108 ;108      ;DEFW 0
VRS_CurESld .equ 110 ;110      ;DEFW 0
VRS_Env_Del .equ 112 ;112      ;DEFB 0
VRS_CurEDel .equ 113 ;113      ;DEFB 0
VRS_Ns_Base .equ 114 ;114      ;DEFB 0
VRS_AddToNs .equ 115 ;115      ;DEFB 0
VRS_AddToEn .equ 116 ;116      ;DEFB 0
VRS_EnvBase .equ 117 ;117      ;DEFW 0
VRS_AYREGS  .equ 119 ;119      ;DEFS 14
                     ;
VRS_CurPos  .equ 133 ;133      ;DEFB 0
VRS_PosSub  .equ 134 ;134      ;DEFB 0
                     ;
VRS_datlen  .equ 135 ;135


VARS1:   .DS VRS_datlen
VARS2:   .DS VRS_datlen
VTBEGIN:
VT_     .EQU VTBEGIN-16
    ; Volume table
    .DB #0x00,#0x00,#0x00,#0x00,#0x00,#0x00,#0x00,#0x00,#0x01,#0x01,#0x01,#0x01,#0x01,#0x01,#0x01,#0x01
    .DB #0x00,#0x00,#0x00,#0x00,#0x01,#0x01,#0x01,#0x01,#0x01,#0x01,#0x01,#0x01,#0x02,#0x02,#0x02,#0x02
    .DB #0x00,#0x00,#0x00,#0x01,#0x01,#0x01,#0x01,#0x01,#0x02,#0x02,#0x02,#0x02,#0x02,#0x03,#0x03,#0x03
    .DB #0x00,#0x00,#0x01,#0x01,#0x01,#0x01,#0x02,#0x02,#0x02,#0x02,#0x03,#0x03,#0x03,#0x03,#0x04,#0x04
    .DB #0x00,#0x00,#0x01,#0x01,#0x01,#0x02,#0x02,#0x02,#0x03,#0x03,#0x03,#0x04,#0x04,#0x04,#0x05,#0x05
    .DB #0x00,#0x00,#0x01,#0x01,#0x02,#0x02,#0x02,#0x03,#0x03,#0x04,#0x04,#0x04,#0x05,#0x05,#0x06,#0x06
    .DB #0x00,#0x00,#0x01,#0x01,#0x02,#0x02,#0x03,#0x03,#0x04,#0x04,#0x05,#0x05,#0x06,#0x06,#0x07,#0x07
    .DB #0x00,#0x01,#0x01,#0x02,#0x02,#0x03,#0x03,#0x04,#0x04,#0x05,#0x05,#0x06,#0x06,#0x07,#0x07,#0x08
    .DB #0x00,#0x01,#0x01,#0x02,#0x02,#0x03,#0x04,#0x04,#0x05,#0x05,#0x06,#0x07,#0x07,#0x08,#0x08,#0x09
    .DB #0x00,#0x01,#0x01,#0x02,#0x03,#0x03,#0x04,#0x05,#0x05,#0x06,#0x07,#0x07,#0x08,#0x09,#0x09,#0x0A
    .DB #0x00,#0x01,#0x01,#0x02,#0x03,#0x04,#0x04,#0x05,#0x06,#0x07,#0x07,#0x08,#0x09,#0x0A,#0x0A,#0x0B
    .DB #0x00,#0x01,#0x02,#0x02,#0x03,#0x04,#0x05,#0x06,#0x06,#0x07,#0x08,#0x09,#0x0A,#0x0A,#0x0B,#0x0C
    .DB #0x00,#0x01,#0x02,#0x03,#0x03,#0x04,#0x05,#0x06,#0x07,#0x08,#0x09,#0x0A,#0x0A,#0x0B,#0x0C,#0x0D
    .DB #0x00,#0x01,#0x02,#0x03,#0x04,#0x05,#0x06,#0x07,#0x07,#0x08,#0x09,#0x0A,#0x0B,#0x0C,#0x0D,#0x0E
    .DB #0x00,#0x01,#0x02,#0x03,#0x04,#0x05,#0x06,#0x07,#0x08,#0x09,#0x0A,#0x0B,#0x0C,#0x0D,#0x0E,#0x0F
                                                                     
;NT_Pointer:  .dw NT_

NT_:
    .if   CPC_FREQ
    ; CPC Note table (it's valid for ST, too)
    .DW 3822,3608,3405,3214,3034,2863,2703,2551,2408,2273,2145,2025
    .DW 1911,1804,1703,1607,1517,1432,1351,1276,1204,1136,1073,1012
    .DW 956,902,851,804,758,716,676,638,602,568,536,506
    .DW 478,451,426,402,379,358,338,319,301,284,268,253
    .DW 239,225,213,201,190,179,169,159,150,142,134,127
    .DW 119,113,106,100,95,89,84,80,75,71,67,63
    .DW 60,56,53,50,47,45,42,40,38,36,34,32
    .DW 30,28,27,25,24,22,21,20,19,18,17,16
;    DEFW 15,14,13,13,12,11,11,10,9,9,8,8
;    DEFW 7,7,7,6,6,6,5,5,5,4,4,4
;    DEFW 4,4,3,3,3,3,3,2,2,2,2,2
;    DEFW 2,2,2,2,1,1,1,1,1,1,1,1
    .ELSE
    ; ZX Note table
;    DEFW 4095,4095,4095,4095,4095,4095,4095,4095,4095,3977,3754,3543
    .DW 3344,3157,2980,2812,2655,2506,2365,2232,2107,1989,1877,1772
    .DW 1672,1578,1490,1406,1327,1253,1182,1116,1053,994,939,886
    .DW 836,789,745,703,664,626,591,558,527,497,469,443
    .DW 418,395,372,352,332,313,296,279,263,249,235,221
    .DW 209,197,186,176,166,157,148,140,132,124,117,111
    .DW 105,99,93,88,83,78,74,70,66,62,59,55
    .DW 52,49,47,44,41,39,37,35,33,31,29,28
    .DW 26,25,23,22,21,20,18,17,16,15,14,13 ; the last 3 should be 16, 15 and 14
;    DEFW 13,12,12,11,10,10,9,9,8,8,7,7
;    DEFW 7,6,6,5,5,5,5,4,4,4,4,3
;    DEFW 3,3,3,3,3,2,2,2,2,2,2,2
    .ENDIF
VAR0END .EQU VT_+16              ; TS_Init zeroes from VARS to VAR0END-1
; ---------------------------------------------------------------------------
