
/***********************************************************************/
/*!
 *  \file       playcityPT3.h
 *  \brief      Header-File for SDCC
 *  \author     Shining \n
 *           
 *
 *  \b Project: Playcity PT3-Support for SDCC \n
 *
 *  \date       20.12.2014 - File created
 *  \date       00.00.2010 - yy
 *  
 *  \version 1.0
 *
 ***********************************************************************/
 
 #ifndef __PLAYCITYPT3_H__
 #define __PLAYCITYPT3_H__




 
/* ++++++++++++++++++++++++++++++++++++++++++++++++++ */
/*! 
 * \brief       Stop playing
 *
 * \b Details:  Stops the music and silences the channels.
 *
 * \param       --
 *
 * \retval      --
 */
/* ++++++++++++++++++++++++++++++++++++++++++++++++++ */
void PC_Stop(void);


/* ++++++++++++++++++++++++++++++++++++++++++++++++++ */
/*! 
 * \brief       Init a song for playing
 *
 * \b Details:  This routine intializes a song in RAM 
 *				for playing. You have to pass the adresses of
 *				the two modules in ram for 6-channel songs. If you
 *				pass the same address for both, you can play a 
 *				3-channel PT3-Song on both chips.
 *
 * \param       firstModule 	Address of the first module in RAM
 * \param       secondModule 	Address of the second module in RAM
 *							
 *
 * \retval      --
 */
/* ++++++++++++++++++++++++++++++++++++++++++++++++++ */
void PC_Init(unsigned char* firstModule, unsigned char* secondModule);


/* ++++++++++++++++++++++++++++++++++++++++++++++++++ */
/*! 
 * \brief       Play the song
 *
 * \b Details:  This routine plays the previously initialized song
 *				has to be called every 20ms (50 Hz). If you call the
 *				player within interrupt you have to define DISABLE_INTERRUPT
 *				in file playcityPT3.asm to 0.
 *
 * \param       firstModule 	Address of the first module in RAM
 * \param       secondModule 	Address of the second module in RAM
 *							
 *
 * \retval      --
 */
/* ++++++++++++++++++++++++++++++++++++++++++++++++++ */
void PC_Play(void);


/* ++++++++++++++++++++++++++++++++++++++++++++++++++ */
/*! 
 * \brief       Get volume form one channel
 *
 * \b Details:  Gets the current volume from one channel.
 *				Remark: Routine checks not for illegal parameters !
 *
 * \param       chip 		which chip (0 or 1)
 * \param       channel 	which channel (0 or 2)
 *
 * \retval      --
 */
/* ++++++++++++++++++++++++++++++++++++++++++++++++++ */ 
unsigned char PC_Get_Volume(unsigned char chip, unsigned char channel);


#endif // __PLAYCITYPT3_H__
